@@ -2,19 +2,19 @@
 # Base class for all VM devices
 #
 # Copyright 2008, 2013 Red Hat, Inc.
-# Cole Robinson <crobinso@redhat.com>
 #
 # This work is licensed under the GNU GPLv2 or later.
 # See the COPYING file in the top-level directory.
 
 import collections
-import logging
 import os
 import re
-import string  # pylint: disable=deprecated-module
+import string
+import textwrap
 
+from .logger import log
 from .xmlapi import XMLAPI
-from . import util
+from . import xmlutil
 
 
 # pylint: disable=protected-access
@@ -26,6 +26,60 @@ _allprops = []
 _seenprops = []
 
 
+class XMLManualAction(object):
+    """
+    Helper class for tracking and performing the user requested manual
+    XML action
+    """
+    ACTION_CREATE = 1
+    ACTION_DELETE = 2
+    ACTION_SET = 3
+
+    xpath_delete = None
+    xpath_create = None
+    xpath_value = None
+    xpath_set = None
+
+    def _process_args(self):
+        def _ret(x, a, v=None):
+            return (x, a, v)
+
+        if self.xpath_delete:
+            return _ret(self.xpath_delete, XMLManualAction.ACTION_DELETE)
+        if self.xpath_create:
+            return _ret(self.xpath_create, XMLManualAction.ACTION_CREATE)
+
+        xpath = self.xpath_set
+        if self.xpath_value:
+            val = self.xpath_value
+        else:
+            if "=" not in str(xpath):
+                raise ValueError(
+                    "%s: Setting xpath must be in the form of XPATH=VALUE" %
+                    xpath)
+            xpath, val = xpath.rsplit("=", 1)
+        return _ret(xpath, XMLManualAction.ACTION_SET, val)
+
+    def perform(self, xmlstate):
+        if (not self.xpath_delete and
+            not self.xpath_create and
+            not self.xpath_value and
+            not self.xpath_set):
+            return
+        xpath, action, value = self._process_args()
+
+        if xpath.startswith("."):
+            xpath = xmlstate.make_abs_xpath(xpath)
+
+        if action == self.ACTION_DELETE:
+            setval = False
+        elif action == self.ACTION_CREATE:
+            setval = True
+        else:
+            setval = value or None
+        xmlstate.xmlapi.set_xpath_content(xpath, setval)
+
+
 class _XMLPropertyCache(object):
     """
     Cache lookup tables mapping classes to their associated
@@ -66,10 +120,11 @@ class _XMLChildList(list):
     This is just to insert a dynamically created add_new() function
     which instantiates and appends a new child object
     """
-    def __init__(self, childclass, copylist, xmlbuilder):
+    def __init__(self, childclass, copylist, xmlbuilder, is_xml=True):
         list.__init__(self)
         self._childclass = childclass
         self._xmlbuilder = xmlbuilder
+        self._is_xml = is_xml
         for i in copylist:
             self.append(i)
 
@@ -77,14 +132,20 @@ class _XMLChildList(list):
         """
         Instantiate a new child object and return it
         """
-        return self._childclass(self._xmlbuilder.conn)
+        args = ()
+        if self._is_xml:
+            args = (self._xmlbuilder.conn,)
+        return self._childclass(*args)
 
     def add_new(self):
         """
         Instantiate a new child object, append it, and return it
         """
         obj = self.new()
-        self._xmlbuilder.add_child(obj)
+        if self._is_xml:
+            self._xmlbuilder.add_child(obj)
+        else:
+            self.append(obj)
         return obj
 
 
@@ -193,7 +254,7 @@ class XMLProperty(_XMLPropertyBase):
         """
         self._xpath = xpath
         if not self._xpath:
-            raise RuntimeError("XMLProperty: xpath must be passed.")
+            raise xmlutil.DevError("XMLProperty: xpath must be passed.")
 
         self._is_bool = is_bool
         self._is_int = is_int
@@ -201,10 +262,11 @@ class XMLProperty(_XMLPropertyBase):
         self._is_onoff = is_onoff
         self._do_abspath = do_abspath
 
-        if sum([int(bool(i)) for i in
+        conflicts = sum(int(bool(i)) for i in
                 [self._is_bool, self._is_int,
-                 self._is_yesno, self._is_onoff]]) > 1:
-            raise RuntimeError("Conflict property converter options.")
+                 self._is_yesno, self._is_onoff])
+        if conflicts > 1:
+            raise xmlutil.DevError("Conflict property converter options.")
 
         self._is_tracked = False
         if _trackprops:
@@ -224,16 +286,30 @@ class XMLProperty(_XMLPropertyBase):
     def _convert_get_value(self, val):
         # pylint: disable=redefined-variable-type
         if self._is_bool:
-            ret = bool(val)
+            return bool(val)
         elif self._is_int and val is not None:
-            intkwargs = {}
-            if "0x" in str(val):
-                intkwargs["base"] = 16
-            ret = int(val, **intkwargs)
-        elif self._is_yesno and val is not None:
-            ret = bool(val == "yes")
-        elif self._is_onoff and val is not None:
-            ret = bool(val == "on")
+            try:
+                intkwargs = {}
+                if "0x" in str(val):
+                    intkwargs["base"] = 16
+                ret = int(val, **intkwargs)
+            except ValueError as e:
+                log.debug("Error converting XML value to int: %s", e)
+                ret = val
+        elif self._is_yesno:
+            if val == "yes":
+                ret = True
+            elif val == "no":
+                ret = False
+            else:
+                ret = val
+        elif self._is_onoff:
+            if val == "on":
+                ret = True
+            elif val == "off":
+                ret = False
+            else:
+                ret = val
         else:
             ret = val
         return ret
@@ -241,10 +317,16 @@ class XMLProperty(_XMLPropertyBase):
     def _convert_set_value(self, val):
         if self._do_abspath and val is not None:
             val = os.path.abspath(val)
-        elif self._is_onoff and val is not None:
-            val = bool(val) and "on" or "off"
-        elif self._is_yesno and val is not None:
-            val = bool(val) and "yes" or "no"
+        elif self._is_onoff:
+            if val is True:
+                val = "on"
+            elif val is False:
+                val = "off"
+        elif self._is_yesno:
+            if val is True:
+                val = "yes"
+            elif val is False:
+                val = "no"
         elif self._is_int and val is not None:
             intkwargs = {}
             if "0x" in str(val):
@@ -261,7 +343,7 @@ class XMLProperty(_XMLPropertyBase):
         propstore = xmlbuilder._propstore
 
         if self.propname in propstore:
-            del(propstore[self.propname])
+            del propstore[self.propname]
         propstore[self.propname] = val
 
     def _nonxml_fget(self, xmlbuilder):
@@ -371,9 +453,13 @@ class _XMLState(object):
         try:
             self.xmlapi = XMLAPI(parsexml)
         except Exception:
-            logging.debug("Error parsing xml=\n%s", parsexml)
+            log.debug("Error parsing xml=\n%s", parsexml)
             raise
 
+        if not self.is_build:
+            # Ensure parsexml has the correct root node
+            self.xmlapi.validate_root_name(self._root_name.split(":")[-1])
+
     def set_relative_object_xpath(self, xpath):
         self._relative_object_xpath = xpath or ""
 
@@ -381,8 +467,6 @@ class _XMLState(object):
         self._parent_xpath = xpath or ""
 
     def _join_xpath(self, x1, x2):
-        if x1.endswith("/"):
-            x1 = x1[:-1]
         if x2.startswith("."):
             x2 = x2[1:]
         return x1 + x2
@@ -413,7 +497,7 @@ class XMLBuilder(object):
     # Name of the root XML element
     XML_NAME = None
 
-    # In some cases, libvirt can incorrectly generate unparseable XML.
+    # In some cases, libvirt can incorrectly generate unparsable XML.
     # These are libvirt bugs, but this allows us to work around it in
     # for specific XML classes.
     #
@@ -425,6 +509,23 @@ class XMLBuilder(object):
     def register_namespace(nsname, uri):
         XMLAPI.register_namespace(nsname, uri)
 
+    @staticmethod
+    def validate_generic_name(name_label, val):
+        # Rather than try and match libvirt's regex, just forbid things we
+        # know don't work
+        forbid = [" "]
+        if not val:
+            # translators: value is a generic object type name
+            raise ValueError(_("A name must be specified for the %s") %
+                    name_label)
+        for c in forbid:
+            if c not in val:
+                continue
+            msg = (_("%(objecttype)s name '%(name)s' can not contain "
+                    "'%(char)s' character.") %
+                    {"objecttype": name_label, "name": val, "char": c})
+            raise ValueError(msg)
+
 
     def __init__(self, conn, parsexml=None,
                  parentxmlstate=None, relative_object_xpath=None):
@@ -439,11 +540,7 @@ class XMLBuilder(object):
         self.conn = conn
 
         if self._XML_SANITIZE:
-            if hasattr(parsexml, 'decode'):
-                parsexml = parsexml.decode("ascii", "ignore").encode("ascii")
-            else:
-                parsexml = parsexml.encode("ascii", "ignore").decode("ascii")
-
+            parsexml = parsexml.encode("ascii", "ignore").decode("ascii")
             parsexml = "".join([c for c in parsexml if c in string.printable])
 
         self._propstore = collections.OrderedDict()
@@ -453,6 +550,8 @@ class XMLBuilder(object):
 
         self._validate_xmlbuilder()
         self._initial_child_parse()
+        self.xml_actions = _XMLChildList(
+                XMLManualAction, [], self, is_xml=False)
 
     def _validate_xmlbuilder(self):
         # This is one time validation we run once per XMLBuilder class
@@ -464,14 +563,15 @@ class XMLBuilder(object):
         childprops = self._all_child_props()
         for key in self._XML_PROP_ORDER:
             if key not in xmlprops and key not in childprops:
-                raise RuntimeError("programming error: key '%s' must be "
-                                   "xml prop or child prop" % key)
+                raise xmlutil.DevError(
+                        "key '%s' must be xml prop or child prop" % key)
 
         childclasses = []
         for childprop in childprops.values():
             if childprop.child_class in childclasses:
-                raise RuntimeError("programming error: can't register "
-                        "duplicate child_classs=%s" % childprop.child_class)
+                raise xmlutil.DevError(
+                        "can't register duplicate child_class=%s" %
+                        childprop.child_class)
             childclasses.append(childprop.child_class)
 
         setattr(self.__class__, cachekey, True)
@@ -519,7 +619,14 @@ class XMLBuilder(object):
         self._add_parse_bits(xmlapi)
         ret = xmlapi.get_xml(self._xmlstate.make_abs_xpath("."))
 
-        if ret and not ret.endswith("\n"):
+        if not ret:
+            return ret
+
+        lastline = ret.rstrip().splitlines()[-1]
+        if not ret.startswith(" ") and lastline.startswith(" "):
+            ret = lastline.split("<")[0] + ret
+
+        if not ret.endswith("\n"):
             ret += "\n"
         return ret
 
@@ -551,7 +658,6 @@ class XMLBuilder(object):
         Validate any set values and raise an exception if there's
         a problem
         """
-        pass
 
     def set_defaults(self, guest):
         """
@@ -598,14 +704,17 @@ class XMLBuilder(object):
 
     def _find_child_prop(self, child_class):
         xmlprops = self._all_child_props()
+        ret = None
         for xmlprop in list(xmlprops.values()):
             if xmlprop.is_single:
                 continue
             if child_class is xmlprop.child_class:
-                return xmlprop
-        raise RuntimeError("programming error: "
-                           "Didn't find child property for child_class=%s" %
-                           child_class)
+                ret = xmlprop
+                break
+        if not ret:
+            raise xmlutil.DevError(
+                "Didn't find child property for child_class=%s" % child_class)
+        return ret
 
     def _set_xpaths(self, parent_xpath, relative_object_xpath=-1):
         """
@@ -615,7 +724,7 @@ class XMLBuilder(object):
         if relative_object_xpath != -1:
             self._xmlstate.set_relative_object_xpath(relative_object_xpath)
         for propname in self._all_child_props():
-            for p in util.listify(getattr(self, propname, [])):
+            for p in xmlutil.listify(getattr(self, propname, [])):
                 p._set_xpaths(self._xmlstate.abs_xpath())
 
     def _set_child_xpaths(self):
@@ -626,7 +735,7 @@ class XMLBuilder(object):
         """
         typecount = {}
         for propname, xmlprop in self._all_child_props().items():
-            for obj in util.listify(getattr(self, propname)):
+            for obj in xmlutil.listify(getattr(self, propname)):
                 idxstr = ""
                 if not xmlprop.is_single:
                     class_type = obj.__class__
@@ -645,7 +754,7 @@ class XMLBuilder(object):
         """
         self._xmlstate.parse(*args, **kwargs)
         for propname in self._all_child_props():
-            for p in util.listify(getattr(self, propname, [])):
+            for p in xmlutil.listify(getattr(self, propname, [])):
                 p._parse_with_children(None, self._xmlstate)
 
     def add_child(self, obj, idx=None):
@@ -669,7 +778,7 @@ class XMLBuilder(object):
             use_xpath = obj._xmlstate.abs_xpath().rsplit("/", 1)[0]
             indent = 2 * obj._xmlstate.abs_xpath().count("/")
             self._xmlstate.xmlapi.node_add_xml(
-                    util.xml_indent(xml, indent), use_xpath)
+                    textwrap.indent(xml, indent * " "), use_xpath)
         obj._parse_with_children(None, self._xmlstate)
 
     def remove_child(self, obj):
@@ -687,6 +796,24 @@ class XMLBuilder(object):
         self._xmlstate.xmlapi.node_force_remove(xpath)
         self._set_child_xpaths()
 
+    def replace_child(self, origobj, newobj):
+        """
+        Replace the origobj child with the newobj. For is_build, this
+        replaces the objects, but for !is_build this only replaces the
+        XML and keeps the object references in place. This is hacky and
+        it's fixable but at time or writing it doesn't matter for
+        our usecases.
+        """
+        if not self._xmlstate.is_build:
+            xpath = origobj.get_xml_id()
+            indent = 2 * xpath.count("/")
+            xml = textwrap.indent(newobj.get_xml(), indent * " ").strip()
+            self._xmlstate.xmlapi.node_replace_xml(xpath, xml)
+        else:
+            origidx = origobj.get_xml_idx()
+            self.remove_child(origobj)
+            self.add_child(newobj, idx=origidx)
+
     def _prop_is_unset(self, propname):
         """
         Return True if the property name has never had a value set
@@ -737,5 +864,8 @@ class XMLBuilder(object):
             if key in xmlprops:
                 xmlprops[key]._set_xml(self, self._propstore[key])
             elif key in childprops:
-                for obj in util.listify(getattr(self, key)):
+                for obj in xmlutil.listify(getattr(self, key)):
                     obj._add_parse_bits(self._xmlstate.xmlapi)
+
+        for manualaction in self.xml_actions:
+            manualaction.perform(self._xmlstate)
